#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <limits.h>
#include <netdb.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define warn(x) perror(x)
#define err(x, y) do { perror(y); exit(x); } while (0)

static void
usage()
{
		fprintf(stderr, "usage:	prtfwd [ -r ] <host>:<port>\n");
		exit(1);
}

#define OH_RETRY	0x1
int
openhost(char *host, int port, int flags)
{
	int fd;
	struct in_addr ia;
	struct sockaddr_in addr;
	struct hostent *hent;
	int len;
	int r;

	if (inet_aton(host, &ia) == 0) {
		if ((hent = gethostbyname(host)) == NULL) {
			herror("gethostbyname");
			return -1;
		}
		ia = *(struct in_addr *)hent->h_addr;
	}

beginsocket:
	if ((fd = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
		warn("socket");
		return -1;
	}

	len = sizeof addr;
	if (getsockname(fd, (struct sockaddr *)&addr, &len) == -1) {
		warn("getsockname");
		return -1;
	}

	addr.sin_addr = ia;
	addr.sin_port = htons(port);
	if ((r = connect(fd, (struct sockaddr *)&addr, len)) == -1 &&
	    (flags & OH_RETRY) &&
	    (errno == ETIMEDOUT || errno == ECONNREFUSED)) {
		close(fd);
		fprintf(stderr, "retrying...\n");
		sleep(1);
		goto beginsocket;
	}
	if (r == -1) {
		warn("connect");
		return -1;
	}

	return fd;
}

int exiting;

void
chldsig(int sig)
{
	int s;

	if (exiting)
		return;

	if (wait3(&s, 0, NULL) == -1)
		exit(1);

	if (WIFEXITED(s) && WEXITSTATUS(s) == 0)
		return;

	exit(1);
}

int fds[2];

void
termsig(int sig)
{
	close(fds[0]);
	close(fds[1]);
	exit(0);
}

int
resolvport(char *str)
{
	int p;
	char *eptr;
	struct servent *a;

	if ((a = getservbyname(str, NULL)) != NULL)
		return ntohs(a->s_port);
	if ((p = strtoul(str, &eptr, 0)), eptr != NULL)
		return p;
	return -1;
}

void
main(int argc, char *argv[])
{
	int fd;
	char *host;
	int port;
	int child;
	char buf[8192];
	int len, i, j;
	struct itimerval to;
	int flags;

	flags = 0;

	argc--; argv++;

	if (argc <= 0)
		usage();
	if (!strcmp(*argv, "-r")) {
		flags |= OH_RETRY;
		argc--; argv++;
	}
	if (argc <= 0)
		usage();

	host = strdup(*argv);
	if (strchr(host, ':') == NULL) {
		fprintf(stderr, "%p\n", strchr(host, ':'));
		usage();
	}
	port = resolvport(strchr(host, ':') + 1);
	*strchr(host,':') = '\0';

	if (port == -1) {
		fprintf(stderr, "failed to resolve port %s.\n",
		    strchr(host, '\0') + 1);
		exit(1);
	}
	if (port <= 0 || port > USHRT_MAX) {
		fprintf(stderr, "%d\n", port);
		usage();
	}

	if ((fd = openhost(host, port, flags)) == -1)
		exit(1);

	exiting = 0;
	fds[0] = STDIN_FILENO;
	fds[1] = fd;

	to.it_interval.tv_sec = 0;
	to.it_interval.tv_usec = 10000;
	to.it_value = to.it_interval;
	/*setitimer(ITIMER_REAL, &to, NULL);*/

	switch ((child = fork())) {
	case -1:
		err(1, "fork failed");
	case 0:
		/* in child */

		signal(SIGTERM, termsig);

		/* read from stdin and write to socket */
		while ((len = read(STDIN_FILENO, buf, sizeof buf)) > 0 ||
		    errno == EINTR)
			for (i = 0, j = 0; j != -1 && i < len; i += j)
				j = write(fd, buf + i, len - i);

		if (len == 0) {
			shutdown(fd, SHUT_WR);

			exit(0);
		}
		exit(1);
		break;
	default:
		/* parent */

		signal(SIGCHLD, chldsig);

		while ((len = read(fd, buf, sizeof buf)) > 0 || errno == EINTR)
			for (i = 0, j = 0; j != -1 && i < len; i += j)
				j = write(STDOUT_FILENO, buf + i, len - i);

		exiting = 1;
		if (len == 0) {
			kill(SIGKILL, child);
			exit(0);
		}
		kill(SIGTERM, child);
		exit(1);
		break;
	}
}
